/** \file mmc.c
	Simple MMC/SD-card functionality
*/

#include <avr/io.h>
#include <inttypes.h>
#include <stdio.h>
#include "mmc.h"

/** Hardware SPI I/O. 
	\param byte Data to send over SPI bus
	\return Received data from SPI bus
*/
uint8_t spi_byte(uint8_t byte)
{
	SPDR = byte;
	while(!(SPSR & (1<<SPIF)))
	{}
	return SPDR;
}

/** Get Token.
	Wait for and return a non-ff token from the MMC/SD card
	\return The received token or 0xFF if timeout
*/
uint8_t mmc_get(void)
{
	uint16_t i = 0xffff;
	uint8_t b = 0xff;

	while ((b == 0xff) && (--i)) 
	{
		b = spi_byte(0xff);
	}
	return b;

}

/** Send a command to the MMC/SD card.
	\param command	Command to send
	\param px	Command parameter 1
	\param py	Command parameter 2
*/
uint8_t mmc_send_command(uint8_t command, uint32_t p)
{
	register union u16convert r;

	MMC_CS_PORT &= ~(1 << MMC_CS);	// enable CS

	spi_byte(0xff);			// dummy byte

	spi_byte(command | 0x40);

	spi_byte(p>>24);	// high byte of param x
	spi_byte(p>>16);	// low byte of param x
	spi_byte(p>>8);	// high byte of param y
	spi_byte(p);	// low byte of param y

	spi_byte(0x95);			// correct CRC for first command in SPI          
							// after that CRC is ignored, so no problem with 
							// always sending 0x95                           
	return mmc_get();			// ignore return byte
}


/** Get Datatoken.
	Wait for and return a data token from the MMC/SD card
	\return The received token or 0xFF if timeout
*/
uint8_t mmc_datatoken(void)
{
	uint16_t i = 0xffff;
	uint8_t b = 0xff;

	while ((b != 0xfe) && (--i)) 
	{
		b = spi_byte(0xff);
	}
	return b;
}

/** Finish Clocking and Release card.
	Send 10 clocks to the MMC/SD card
 	and release the CS line 
*/
void mmc_clock_and_release(void)
{

	// SD cards require at least 8 final clocks
	while(spi_byte(0xFF)!=0xFF);

    MMC_CS_PORT |= (1 << MMC_CS);	// release CS
}

/** Read MMC/SD sector.
 	Read a single 512 byte sector from the MMC/SD card
	\param lba	Logical sectornumber to read
	\param buffer	Pointer to buffer for received data
	\return 0 on success, -1 on error
*/
int mmc_readsector(uint32_t lba, uint8_t *buffer)
{
	uint16_t i;

	// send read command and logical sector address
	mmc_send_command(17,lba<<9);

	if (mmc_datatoken() != 0xfe)	// if no valid token
	{
	    mmc_clock_and_release();	// cleanup and	
   		return -1;					// return error code
	}

	for (i=0;i<256;i++)				// read sector data
    	*buffer++ = spi_byte(0xff);
	for (i=0;i<256;i++)
		spi_byte(0xff);

	spi_byte(0xff);					// ignore dummy checksum
	spi_byte(0xff);					// ignore dummy checksum

    mmc_clock_and_release();		// cleanup

	return 0;						// return success		
}

/** Write MMC/SD sector.
 	Write a single 512 byte sector to the MMC/SD card
	\param lba	Logical sectornumber to be written
	\param buffer	Pointer to buffer for data to be written
	\return 0 on success, -1 on error
*/
int mmc_writesector(uint32_t lba, uint8_t *buffer)
{
	uint16_t i, j;
	uint8_t c;
	
	
	// send write command and logical sector address
	c = mmc_send_command(24,lba<<9);
	if (c)
	{
		mmc_clock_and_release();	// cleanup and	
   		return -c;					// return error code
	}
	
	spi_byte(0xff);
	//spi_byte(0xff);
	spi_byte(0xfe);

	for (i=0;i<512;i++)	{ // write sector data
		if (!buffer)
			break;
    	spi_byte(*(buffer++));
	}
	for (j=i;j<512;j++)				// padding
		spi_byte(0x00);	

	spi_byte(0xff);					// ignore dummy checksum
	spi_byte(0xff);					// ignore dummy checksum
	
	c = spi_byte(0xff);
	c &= 0x1F; 	// 0x1F = 0b.0001.1111;
	if (c != 0x05) { // 0x05 = 0b.0000.0101
		mmc_clock_and_release();		// cleanup
		return -2;
	}
    mmc_clock_and_release();		// cleanup
	return 0;						// return success		
}

/** Init MMC/SD card.
	Initialize I/O ports for the MMC/SD interface and 
	send init commands to the MMC/SD card
	\return 0 on success, other values on error 
*/
uint8_t mmc_init(void)
{
	int i;
	uint8_t c=1;


	// setup I/O ports 

	SPI_PORT &= ~((1 << MMC_SCK) | (1 << MMC_MOSI));	// low bits
	SPI_PORT |= (1 << MMC_MISO);						// high bits
	SPI_DDR  |= (1<<MMC_SCK) | (1<<MMC_MOSI);			// direction


	MMC_CS_PORT |= (1 << MMC_CS);	// Initial level is high	
	MMC_CS_DIR  |= (1 << MMC_CS);	// Direction is output


	// also need to set SS as output
#if defined(__AVR_ATmega8__)
	// is already set as CS, but we set it again to accomodate for other boards 
	SPI_DDR |= (1<<2);
#else
	SPI_DDR |= (1<<0);			// assume it's bit0 (mega128, portB and others)
#endif

	SPCR = (1<<MSTR)|(1<<SPE);	// enable SPI interface
	SPSR = 1;					// set double speed	

	for(i=0;i<10;i++)			// send 80 clocks while card power stabilizes
		spi_byte(0xff);

	// send CMD0 - reset card

	if (mmc_send_command(0,0) != 1)			// if no valid response code
	{
	   mmc_clock_and_release();
	   return 1;  				// card cannot be detected
	}

	//
	// send CMD1 until we get a 0 back, indicating card is done initializing 
	//
	i = 0xffff;						// max timeout
	while ((c != 0) && (--i))	// wait for it
	{
	     c = mmc_send_command(1,0);	// send CMD1 - activate card init
	}

    mmc_clock_and_release();		// clean up

	if (i == 0)						// if we timed out above
	   return 2;					// return failure code

	return 0;
}
